# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import codecs
import contextlib
import nfoview
import sys
import traceback
import urllib.parse
import webbrowser

from gi.repository import Gdk
from gi.repository import Gtk
from gi.repository import Pango

def affirm(value):
    if not value:
        raise nfoview.AffirmationError(f"Not True: {value!r}")

def apply_style(widget):
    name = nfoview.conf.color_scheme
    scheme = nfoview.schemes.get(name, "default")
    font_desc = Pango.FontDescription(nfoview.conf.font)
    css = """
    .nfoview-text-view, .nfoview-text-view text {{
        background-color: {bg};
        color: {fg};
        font-family: "{family}", monospace;
        font-size: {size}px;
        font-weight: {weight};
    }}""".format(
        bg=scheme.background,
        fg=scheme.foreground,
        family=font_desc.get_family().split(",")[0].strip('"'),
        size=int(round(font_desc.get_size() / Pango.SCALE)),
        # Round weight to hundreds to work around CSS errors
        # with weird weights such as Unscii's 101.
        weight=round(font_desc.get_weight(), -2),
    )
    css = css.replace("font-size: 0px;", "")
    css = css.replace("font-weight: 0;", "")
    css = "\n".join(filter(lambda x: x.strip(), css.splitlines()))
    provider = Gtk.CssProvider()
    try:
        # The call signature of 'load_from_data' seems to have changed
        # in some GTK version. Also, the whole function is deprecated
        # and since GTK 4.12 we should use 'load_from_string'.
        provider.load_from_data(css, -1)
    except Exception:
        provider.load_from_data(bytes(css.encode()))
    style = widget.get_style_context()
    style.add_class("nfoview-text-view")
    display = Gdk.Display.get_default()
    priority = Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
    style.add_provider_for_display(display, provider, priority)

def connect(observer, observable, signal, *args):
    # If observable is a string, it should be an attribute of observer.
    # If observable is not a string it should be the same as observer.
    method_name = signal.replace("-", "_").replace("::", "_")
    if observer is not observable:
        method_name = "_".join((observable, method_name))
    method_name = f"_on_{method_name}".replace("__", "_")
    if not hasattr(observer, method_name):
        method_name = method_name[1:]
    method = getattr(observer, method_name)
    if observer is not observable:
        observable = getattr(observer, observable)
    return observable.connect(signal, method, *args)

def detect_encoding(path, default="cp437"):
    with open(path, "rb") as f:
        line = f.readline()
    if (line.startswith(codecs.BOM_UTF32_BE) and
        is_valid_encoding("utf_32_be")):
        return "utf_32_be"
    if (line.startswith(codecs.BOM_UTF32_LE) and
        is_valid_encoding("utf_32_le")):
        return "utf_32_le"
    if (line.startswith(codecs.BOM_UTF8) and
        is_valid_encoding("utf_8_sig")):
        return "utf_8_sig"
    if (line.startswith(codecs.BOM_UTF16_BE) and
        is_valid_encoding("utf_16_be")):
        return "utf_16_be"
    if (line.startswith(codecs.BOM_UTF16_LE) and
        is_valid_encoding("utf_16_le")):
        return "utf_16_le"
    return default

def get_max_text_view_size():
    max_chars = nfoview.conf.text_view_max_chars
    max_lines = nfoview.conf.text_view_max_lines
    max_text = "\n".join(("x" * max_chars,) * max_lines)
    return get_text_view_size(max_text)

def get_monitor():
    display = Gdk.Display.get_default()
    for monitor in display.get_monitors():
        if monitor is not None:
            return monitor

def get_screen_size(monitor=None):
    monitor = monitor or get_monitor()
    rect = monitor.get_geometry()
    return rect.width, rect.height

def get_text_view_size(text):
    label = Gtk.Label()
    apply_style(label)
    label.set_text(text)
    label.show()
    width = label.measure(Gtk.Orientation.HORIZONTAL, -1)
    height = label.measure(Gtk.Orientation.VERTICAL, -1)
    return width.natural, height.natural

def hex_to_rgba(string):
    rgba = Gdk.RGBA()
    success = rgba.parse(string)
    if success: return rgba
    raise ValueError(f"Parsing {string!r} failed")

def is_valid_encoding(encoding):
    try:
        return codecs.lookup(encoding)
    except LookupError:
        return False

def lookup_color(name, fallback):
    entry = Gtk.Entry()
    entry.show()
    style = entry.get_style_context()
    found, color = style.lookup_color(name)
    if found:
        return rgba_to_hex(color)
    return fallback

def rgba_to_hex(color):
    return "#{:02x}{:02x}{:02x}".format(
        int(color.red   * 255),
        int(color.green * 255),
        int(color.blue  * 255),
    )

def show_uri(uri):
    try:
        return Gtk.show_uri(None, uri, Gdk.CURRENT_TIME)
    except Exception:
        # Gtk.show_uri fails on Windows and some misconfigured installations.
        # GError: No application is registered as handling this file
        # Gtk.show_uri: Operation not supported
        if uri.startswith(("http://", "https://")):
            return webbrowser.open(uri)
        raise # Exception

@contextlib.contextmanager
def silent(*exceptions, tb=False):
    try:
        yield
    except exceptions:
        if tb: traceback.print_exc()

def uri_to_path(uri):
    uri = urllib.parse.unquote(uri)
    if sys.platform == "win32":
        path = urllib.parse.urlsplit(uri)[2]
        while path.startswith("/"):
            path = path[1:]
        return path.replace("/", "\\")
    return urllib.parse.urlsplit(uri)[2]



# -*- coding: utf-8 -*-

# Copyright (C) 2008 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import nfoview

from gi.repository import GObject
from gi.repository import Gtk
from nfoview.i18n import _

def boxwrap(widget):
    # Needed to get widget natural-size left-aligned in grid.
    box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=0)
    box.append(widget)
    return box

def build_label(text):
    label = Gtk.Label.new(text)
    label.add_css_class("dim-label")
    label.set_xalign(1)
    return label

class PreferencesDialog(Gtk.Dialog):

    def __init__(self, parent):
        GObject.GObject.__init__(self, use_header_bar=True)
        self.set_resizable(False)
        self.set_title(_("Preferences"))
        self.set_transient_for(parent)
        grid = Gtk.Grid()
        grid.set_column_spacing(18)
        grid.set_margin_bottom(18)
        grid.set_margin_end(18)
        grid.set_margin_start(18)
        grid.set_margin_top(18)
        grid.set_row_homogeneous(True)
        grid.set_row_spacing(12)

        # Font
        self._font_label = build_label(_("Font"))
        self._font_button = Gtk.FontButton()
        def monospace(family, *args, **kwargs):
            return family.is_monospace()
        self._font_button.set_filter_func(monospace, None)
        self._font_button.set_font(nfoview.conf.font)
        self._font_button.connect("font-set", self._on_font_button_font_set)
        grid.attach(self._font_label, 0, 0, 1, 1)
        grid.attach(boxwrap(self._font_button), 1, 0, 1, 1)

        # Line-spacing
        self._line_spacing_label = build_label(_("Line-spacing"))
        self._line_spacing_spin = Gtk.SpinButton.new_with_range(-99, 99, 1)
        self._line_spacing_spin.set_value(nfoview.conf.pixels_above_lines)
        self._line_spacing_spin.connect("value-changed", self._on_line_spacing_spin_value_changed)
        grid.attach(self._line_spacing_label, 0, 1, 1, 1)
        grid.attach(boxwrap(self._line_spacing_spin), 1, 1, 1, 1)

        # Color scheme
        self._scheme_label = build_label(_("Color scheme"))
        self._scheme_combo = Gtk.ComboBoxText.new()
        for i, scheme in enumerate(nfoview.schemes.get_all()):
            self._scheme_combo.append_text(scheme.label)
            if scheme.name == nfoview.conf.color_scheme:
                self._scheme_combo.set_active(i)
        self._scheme_combo.connect("changed", self._on_scheme_combo_changed)
        grid.attach(self._scheme_label, 0, 2, 1, 1)
        grid.attach(boxwrap(self._scheme_combo), 1, 2, 1, 1)

        # Foreground
        self._fg_color_label = build_label(_("Foreground"))
        self._fg_color_button = Gtk.ColorButton()
        color = nfoview.util.hex_to_rgba(nfoview.conf.foreground_color)
        self._fg_color_button.set_rgba(color)
        self._fg_color_button.connect("color-set", self._on_fg_color_button_color_set)
        grid.attach(self._fg_color_label, 0, 3, 1, 1)
        grid.attach(boxwrap(self._fg_color_button), 1, 3, 1, 1)

        # Background
        self._bg_color_label = build_label(_("Background"))
        self._bg_color_button = Gtk.ColorButton()
        color = nfoview.util.hex_to_rgba(nfoview.conf.background_color)
        self._bg_color_button.set_rgba(color)
        self._bg_color_button.connect("color-set", self._on_bg_color_button_color_set)
        grid.attach(self._bg_color_label, 0, 4, 1, 1)
        grid.attach(boxwrap(self._bg_color_button), 1, 4, 1, 1)

        # Link
        self._link_color_label = build_label(_("Link"))
        self._link_color_button = Gtk.ColorButton()
        color = nfoview.util.hex_to_rgba(nfoview.conf.link_color)
        self._link_color_button.set_rgba(color)
        self._link_color_button.connect("color-set", self._on_link_color_button_color_set)
        grid.attach(self._link_color_label, 0, 5, 1, 1)
        grid.attach(boxwrap(self._link_color_button), 1, 5, 1, 1)

        # Visited link
        self._vlink_color_label = build_label(_("Visited link"))
        self._vlink_color_button = Gtk.ColorButton()
        color = nfoview.util.hex_to_rgba(nfoview.conf.visited_link_color)
        self._vlink_color_button.set_rgba(color)
        self._vlink_color_button.connect("color-set", self._on_vlink_color_button_color_set)
        grid.attach(self._vlink_color_label, 0, 6, 1, 1)
        grid.attach(boxwrap(self._vlink_color_button), 1, 6, 1, 1)

        # Export Scaling
        self._export_scale_label = build_label(_("Export to PNG scaling"))
        self._export_scale_spin = Gtk.SpinButton.new_with_range(1, 5, 0.5)
        self._export_scale_spin.set_value(nfoview.conf.export_scale)
        self._export_scale_spin.connect("value-changed", self._on_export_scaling_spin_value_changed)
        grid.attach(self._export_scale_label, 0, 7, 1, 1)
        grid.attach(boxwrap(self._export_scale_spin), 1, 7, 1, 1)

        self._update_sensitivities()
        self.set_child(grid)
        self.show()

    def _get_windows(self):
        return nfoview.app.get_windows() if hasattr(nfoview, "app") else []

    def _on_bg_color_button_color_set(self, color_button):
        color = color_button.get_rgba()
        color = nfoview.util.rgba_to_hex(color)
        nfoview.conf.background_color = color
        scheme = nfoview.schemes.get("custom")
        scheme.background = color
        for window in self._get_windows():
            window.view.update_style()

    def _on_font_button_font_set(self, font_button):
        nfoview.conf.font = font_button.get_font()
        for window in self._get_windows():
            window.view.update_style()

    def _on_fg_color_button_color_set(self, color_button):
        color = color_button.get_rgba()
        color = nfoview.util.rgba_to_hex(color)
        nfoview.conf.foreground_color = color
        scheme = nfoview.schemes.get("custom")
        scheme.foreground = color
        for window in self._get_windows():
            window.view.update_style()

    def _on_line_spacing_spin_value_changed(self, spin_button):
        pixels = spin_button.get_value_as_int()
        nfoview.conf.pixels_above_lines = pixels
        for window in self._get_windows():
            window.view.set_pixels_above_lines(pixels)

    def _on_link_color_button_color_set(self, color_button):
        color = color_button.get_rgba()
        color = nfoview.util.rgba_to_hex(color)
        nfoview.conf.link_color = color
        scheme = nfoview.schemes.get("custom")
        scheme.link = color
        for window in self._get_windows():
            window.view.update_style()

    def _on_scheme_combo_changed(self, combo_box):
        index = combo_box.get_active()
        scheme = nfoview.schemes.get_all()[index]
        nfoview.conf.color_scheme = scheme.name
        self._update_color_buttons(scheme)
        for window in self._get_windows():
            window.view.update_style()
        self._update_sensitivities()

    def _on_vlink_color_button_color_set(self, color_button):
        color = color_button.get_rgba()
        color = nfoview.util.rgba_to_hex(color)
        nfoview.conf.visited_link_color = color
        scheme = nfoview.schemes.get("custom")
        scheme.vlink = color
        for window in self._get_windows():
            window.view.update_style()

    def _on_export_scaling_spin_value_changed(self, spin_button):
        nfoview.conf.export_scale = spin_button.get_value()

    def _update_color_buttons(self, scheme):
        rgba = nfoview.util.hex_to_rgba
        self._bg_color_button.set_rgba(rgba(scheme.background))
        self._fg_color_button.set_rgba(rgba(scheme.foreground))
        self._link_color_button.set_rgba(rgba(scheme.link))
        self._vlink_color_button.set_rgba(rgba(scheme.visited_link))

    def _update_sensitivities(self):
        sensitive = (nfoview.conf.color_scheme == "custom")
        self._bg_color_button.set_sensitive(sensitive)
        self._fg_color_button.set_sensitive(sensitive)
        self._link_color_button.set_sensitive(sensitive)
        self._vlink_color_button.set_sensitive(sensitive)



# -*- coding: utf-8 -*-

# Copyright (C) 2015 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import nfoview

__all__ = (
    "AboutAction",
    "CloseAction",
    "ExportImageAction",
    "OpenAction",
    "PreferencesAction",
    "QuitAction",
    "WrapLinesAction",
)

class AboutAction(nfoview.Action):
    def __init__(self):
        nfoview.Action.__init__(self, "about")

class CloseAction(nfoview.Action):
    def __init__(self):
        nfoview.Action.__init__(self, "close")
        self.accelerators = ["<Control>W", "Escape"]

class ExportImageAction(nfoview.Action):
    def __init__(self):
        nfoview.Action.__init__(self, "export-image")
        self.accelerators = ["<Control>E"]
    def _affirm_doable(self, window):
        nfoview.util.affirm(window.path is not None)
        nfoview.util.affirm(window.view is not None)
        nfoview.util.affirm(window.view.get_sensitive())

class OpenAction(nfoview.Action):
    def __init__(self):
        nfoview.Action.__init__(self, "open")
        self.accelerators = ["<Control>O"]

class PreferencesAction(nfoview.Action):
    def __init__(self):
        nfoview.Action.__init__(self, "preferences")

class QuitAction(nfoview.Action):
    def __init__(self):
        nfoview.Action.__init__(self, "quit")
        self.accelerators = ["<Control>Q"]

class WrapLinesAction(nfoview.ToggleAction):
    def __new__(cls):
        action = nfoview.ToggleAction.new("wrap-lines")
        action.__class__ = cls
        return action
    def __init__(self):
        nfoview.Action.__init__(self, "wrap-lines")
        self.accelerators = ["<Control>R"]
    def _affirm_doable(self, window):
        nfoview.util.affirm(window.view is not None)
        nfoview.util.affirm(window.view.get_sensitive())
        nfoview.util.affirm(window.view.get_text())



# -*- coding: utf-8 -*-

# Copyright (C) 2013 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GObject
from gi.repository import Gtk
from nfoview.i18n  import _

# XXX: Gtk.FileChooserNative fails on GTK 4.10 with
# "The folder contents could not be displayed"
# "Operation was cancelled"
class ExportImageDialog(Gtk.FileChooserDialog):

    def __init__(self, parent):
        GObject.GObject.__init__(self)
        self.set_action(Gtk.FileChooserAction.SAVE)
        self.set_select_multiple(False)
        self.set_title(_("Export Image"))
        self.set_transient_for(parent)
        self.add_button(_("_Cancel"), Gtk.ResponseType.CANCEL)
        self.add_button(_("_Save"), Gtk.ResponseType.ACCEPT)
        self.set_default_response(Gtk.ResponseType.ACCEPT)



# -*- coding: utf-8 -*-

# Copyright (C) 2008 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import nfoview

from gi.repository import GObject
from gi.repository import Gtk
from nfoview.i18n  import _

class AboutDialog(Gtk.AboutDialog):

    def __init__(self, parent):
        GObject.GObject.__init__(self)
        self.set_authors(("Osmo Salomaa <otsaloma@iki.fi>",))
        self.set_comments(_("Viewer for NFO files"))
        self.set_copyright("Copyright © 2005–2025 Osmo Salomaa")
        self.set_license_type(Gtk.License.GPL_3_0)
        self.set_logo_icon_name("io.otsaloma.nfoview")
        self.set_title(_("About NFO Viewer"))
        self.set_transient_for(parent)
        self.set_version(nfoview.__version__)
        self.set_website("https://otsaloma.io/nfoview")
        self.set_website_label("https://otsaloma.io/nfoview")

        # TRANSLATORS: The application name "NFO Viewer" has been intentionally
        # marked as translatable. If you manage to translate the name in a
        # fluent manner, without changing the meaning, you may use that
        # translation at your discretion.
        self.set_program_name(_("NFO Viewer"))

        # TRANSLATORS: This is a special message that shouldn't be translated
        # literally. It is used in the about dialog to give credits to the
        # translators. Thus, you should translate it to your name and email
        # address. You can also include other translators who have contributed
        # to this translation; in that case, please write them on separate
        # lines seperated by newlines (\n).
        self.set_translator_credits(_("translator-credits"))



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import atexit
import os
import tempfile
import time

from gi.repository import GLib
from pathlib import Path

class TestCase:

    def main_loop(self, window):
        main_context = GLib.MainContext.default()
        while window.get_visible():
            while main_context.pending():
                main_context.iteration()
            time.sleep(0.01)

    def new_nfo_file(self):
        handle, path = tempfile.mkstemp()
        with os.fdopen(handle, "w") as f:
            f.write(self.new_nfo_text())
        atexit.register(os.remove, path)
        return Path(path)

    def new_nfo_text(self):
        return "\n".join((
            "qwertyuiop asdfghjkl zxcvbnm",
            "https://otsaloma.io/nfoview",
        )) + "\n"

    def setUp(self):
        self.setup_method(None)

    def setup_method(self, method):
        pass

    def tearDown(self):
        self.teardown_method(None)

    def teardown_method(self, method):
        pass

    def test___init__(self):
        # Make sure that setup_method is always run.
        pass



# -*- coding: utf-8 -*-

# Copyright (C) 2008 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GObject
from gi.repository import Gtk
from nfoview.i18n  import _

# XXX: Gtk.FileChooserNative fails on GTK 4.10 with
# "The folder contents could not be displayed"
# "Operation was cancelled"
class OpenDialog(Gtk.FileChooserDialog):

    def __init__(self, parent):
        GObject.GObject.__init__(self)
        self.set_action(Gtk.FileChooserAction.OPEN)
        self.set_select_multiple(True)
        self.set_title(_("Open"))
        self.set_transient_for(parent)
        self.add_button(_("_Cancel"), Gtk.ResponseType.CANCEL)
        self.add_button(_("_Open"), Gtk.ResponseType.ACCEPT)
        self.set_default_response(Gtk.ResponseType.ACCEPT)
        self._add_filters()

    def _add_filters(self):
        file_filter = Gtk.FileFilter()
        file_filter.set_name(_("All files"))
        file_filter.add_pattern("*")
        self.add_filter(file_filter)
        file_filter = Gtk.FileFilter()
        file_filter.set_name(_("NFO files (*.nfo)"))
        file_filter.add_pattern("*.[Nn][Ff][Oo]")
        self.add_filter(file_filter)
        self.set_filter(file_filter)



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import nfoview
import re

from gi.repository import Gdk
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Pango

class TextView(Gtk.TextView):

    def __init__(self):
        GObject.GObject.__init__(self)
        self._link_tags = []
        self._visited_link_tags = []
        self.set_bottom_margin(6)
        self.set_cursor_visible(False)
        self.set_editable(False)
        self.set_left_margin(6)
        self.set_pixels_above_lines(nfoview.conf.pixels_above_lines)
        self.set_pixels_below_lines(nfoview.conf.pixels_below_lines)
        self.set_right_margin(6)
        self.set_top_margin(6)
        self.set_wrap_mode(Gtk.WrapMode.NONE)
        controller = Gtk.EventControllerMotion()
        controller.set_propagation_phase(Gtk.PropagationPhase.CAPTURE)
        self.add_controller(controller)
        controller.connect("motion", self._on_motion)
        gesture = Gtk.GestureClick()
        self.add_controller(gesture)
        gesture.connect("pressed", self._on_pressed)
        self.update_style()

    def get_text(self):
        text_buffer = self.get_buffer()
        start, end = text_buffer.get_bounds()
        return text_buffer.get_text(start, end, False)

    def _insert_text(self, text):
        text_buffer = self.get_buffer()
        itr = text_buffer.get_end_iter()
        text_buffer.insert(itr, text)

    def _insert_url(self, url):
        text_buffer = self.get_buffer()
        tag = text_buffer.create_tag(None)
        tag.props.underline = Pango.Underline.SINGLE
        tag.nfoview_url = url
        itr = text_buffer.get_end_iter()
        text_buffer.insert_with_tags(itr, url, tag)
        self._link_tags.append(tag)

    def _on_motion(self, controller, x, y, user_data=None):
        window = Gtk.TextWindowType.WIDGET
        x, y = self.window_to_buffer_coords(window, x, y)
        if iter := self.get_iter_at_location(x, y):
            for tag in iter.iter.get_tags():
                if hasattr(tag, "nfoview_url"):
                    return self.set_cursor(Gdk.Cursor.new_from_name("pointer"))
        self.set_cursor(Gdk.Cursor.new_from_name("default"))

    def _on_pressed(self, gesture, n_press, x, y, user_data=None):
        text_buffer = self.get_buffer()
        if text_buffer.get_selection_bounds(): return
        window = Gtk.TextWindowType.WIDGET
        x, y = self.window_to_buffer_coords(window, x, y)
        if iter := self.get_iter_at_location(x, y):
            for tag in iter.iter.get_tags():
                if hasattr(tag, "nfoview_url"):
                    nfoview.util.show_uri(tag.nfoview_url)
                    if tag in self._link_tags:
                        self._link_tags.remove(tag)
                        self._visited_link_tags.append(tag)
                        self.update_style()

    def set_text(self, text):
        URL = r"(\w+://(\S+\.)?\S+|www\.\S+)\.[\w\-.~:/?#\[\]@!$&'()*+,;=%]+"
        text_buffer = self.get_buffer()
        bounds = text_buffer.get_bounds()
        text_buffer.delete(*bounds)
        self._link_tags = []
        self._visited_link_tags = []
        for line in text.splitlines():
            i = 0
            for match in re.finditer(URL, line):
                a, z = match.span()
                self._insert_text(line[i:a])
                self._insert_url(line[a:z])
                i = z
            self._insert_text(line[i:])
            self._insert_text("\n")
        self.update_style()

    def update_style(self):
        nfoview.util.apply_style(self)
        name = nfoview.conf.color_scheme
        scheme = nfoview.schemes.get(name, "default")
        for tag in self._link_tags:
            color = nfoview.util.hex_to_rgba(scheme.link)
            tag.props.foreground_rgba = color
        for tag in self._visited_link_tags:
            color = nfoview.util.hex_to_rgba(scheme.visited_link)
            tag.props.foreground_rgba = color



# -*- coding: utf-8 -*-

# Copyright (C) 2008 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import nfoview
import re

from pathlib import Path

DEFAULTS = {
    "background_color": "#ffffff",
    "color_scheme": "default",
    "export_scale": 1.0,
    "font": "Cascadia Mono 12",
    "foreground_color": "#2e3436",
    "link_color": "#2a76c6",
    "pixels_above_lines": 0,
    "pixels_below_lines": 0,
    "text_view_max_chars": 160,
    "text_view_max_lines": 45,
    "version": "",
    "visited_link_color": "#215d9c",
}

class ConfigurationStore:

    path = Path(nfoview.CONFIG_HOME_DIR) / "nfoview.conf"

    def __init__(self, read=False):
        self.restore_defaults()
        if read: self.read()

    def read(self):
        if not self.path.exists(): return
        entries = self.path.read_text().splitlines()
        entries = dict(
            re.split(" *= *", x.strip(), maxsplit=1)
            for x in entries
            if not x.startswith("#") and "=" in x
        )
        for name in set(DEFAULTS) & set(entries):
            decode = type(DEFAULTS[name])
            setattr(self, name, decode(entries[name]))
        self.version = nfoview.__version__

    def restore_defaults(self):
        for name in DEFAULTS:
            setattr(self, name, DEFAULTS[name])
        self.version = nfoview.__version__

    def write(self):
        self.path.parent.mkdir(parents=True, exist_ok=True)
        with open(self.path, "w") as f:
            for name in sorted(DEFAULTS):
                text = f"{name} = {getattr(self, name)!s}"
                if getattr(self, name) == DEFAULTS[name]:
                    # Comment out options at default value.
                    text = f"# {text}"
                f.write(text + "\n")



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import cairo
import nfoview
import textwrap

from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Pango
from gi.repository import PangoCairo
from nfoview.i18n  import _
from pathlib import Path

class Window(Gtk.ApplicationWindow):

    def __init__(self, path=None):
        GObject.GObject.__init__(self)
        self.path = Path(path) if path else None
        self.view = nfoview.TextView()
        self._about_dialog = None
        self._prefs_dialog = None
        self._init_properties()
        self._init_titlebar()
        self._init_contents()
        self._init_actions()
        self.open_file(self.path)
        self.resize_to_text()
        self._update_actions_enabled()

    def _init_actions(self):
        for name in nfoview.actions.__all__:
            action = getattr(nfoview.actions, name)()
            if hasattr(nfoview, "app"):
                nfoview.app.set_accels_for_action(
                    f"win.{action.props.name}", action.accelerators)
            callback = f"_on_{action.props.name}_activate".replace("-", "_")
            action.connect("activate", getattr(self, callback))
            self.add_action(action)

    def _init_contents(self):
        scroller = Gtk.ScrolledWindow()
        scroller.set_policy(*((Gtk.PolicyType.AUTOMATIC,)*2))
        scroller.set_hexpand(True)
        scroller.set_vexpand(True)
        scroller.set_child(self.view)
        self.set_child(scroller)

    def _init_properties(self):
        self.set_title(_("NFO Viewer"))
        self.set_icon_name("io.otsaloma.nfoview")
        Gtk.Window.set_default_icon_name("io.otsaloma.nfoview")
        self.connect("close-request", self._on_close_request)
        target = Gtk.DropTarget.new(Gio.File, Gdk.DragAction.COPY)
        target.connect("drop", self._on_drag_drop)
        self.view.add_controller(target)

    def _init_titlebar(self):
        header = Gtk.HeaderBar()
        menu_button = Gtk.MenuButton()
        menu_button.set_direction(Gtk.ArrowType.NONE)
        path = nfoview.DATA_DIR / "menu.ui"
        builder = Gtk.Builder.new_from_file(str(path))
        menu = builder.get_object("menu")
        menu_button.set_menu_model(menu)
        header.pack_start(menu_button)
        self.set_titlebar(header)

    def _on_drag_drop(self, target, value, x, y, user_data=None):
        path = value.get_path()
        if self.path is None:
            return self.open_file(path)
        nfoview.app.open_window(path)

    def _on_about_activate(self, *args):
        if self._about_dialog:
            return self._about_dialog.present()
        def on_close_request(dialog, *args, **kwargs):
            dialog.hide()
            return True
        self._about_dialog = nfoview.AboutDialog(self)
        self._about_dialog.connect("close-request", on_close_request)
        self._about_dialog.show()

    def _on_close_activate(self, *args):
        self.destroy()
        if hasattr(nfoview, "app"):
            nfoview.app.remove_window(self)

    def _on_close_request(self, *args):
        self.destroy()
        if hasattr(nfoview, "app"):
            nfoview.app.remove_window(self)

    def _on_export_image_activate(self, *args):
        dialog = nfoview.ExportImageDialog(self)
        directory = Gio.File.new_for_path(str(self.path.parent))
        dialog.set_current_folder(directory)
        dialog.set_current_name(f"{self.path.name}.png")
        dialog.connect("response", self._on_export_image_activate_response)
        dialog.show()

    def _on_export_image_activate_response(self, dialog, response):
        path = dialog.get_file().get_path()
        dialog.destroy()
        if response not in (
            Gtk.ResponseType.ACCEPT,
            Gtk.ResponseType.OK,
        ): return
        if not path: return
        # Get text buffer and full text
        buffer = self.view.get_buffer()
        start_iter = buffer.get_start_iter()
        end_iter = buffer.get_end_iter()
        text = buffer.get_text(start_iter, end_iter, True)
        # Use Pango to measure
        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, 1, 1)
        cr = cairo.Context(surface)
        layout = PangoCairo.create_layout(cr)
        font_desc = Pango.FontDescription(nfoview.conf.font)
        layout.set_font_description(font_desc)
        layout.set_text(text, -1)
        width, height = layout.get_pixel_size()
        # Create a surface to hold the full layout
        scale = nfoview.conf.export_scale
        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, (int)(scale * width), (int)(scale * height))
        surface.set_device_scale(scale, scale)
        cr = cairo.Context(surface)
        # White background
        cr.set_source_rgb(1, 1, 1)
        cr.paint()
        # Render text
        PangoCairo.update_layout(cr, layout)
        cr.set_source_rgb(0, 0, 0)
        PangoCairo.show_layout(cr, layout)
        surface.write_to_png(path)

    def _on_open_activate(self, *args):
        dialog = nfoview.OpenDialog(self)
        if self.path is not None:
            directory = Gio.File.new_for_path(str(self.path.parent))
            dialog.set_current_folder(directory)
        dialog.connect("response", self._on_open_activate_response)
        dialog.show()

    def _on_open_activate_response(self, dialog, response):
        paths = [x.get_path() for x in dialog.get_files()]
        dialog.destroy()
        if response not in (
            Gtk.ResponseType.ACCEPT,
            Gtk.ResponseType.OK,
        ): return
        for path in paths:
            if self.path is None:
                self.open_file(path)
            elif hasattr(nfoview, "app"):
                nfoview.app.open_window(path)

    def _on_quit_activate(self, *args):
        if hasattr(nfoview, "app"):
            nfoview.app.quit()

    def _on_preferences_activate(self, *args):
        if self._prefs_dialog:
            return self._prefs_dialog.present()
        def on_close_request(dialog, *args, **kwargs):
            dialog.hide()
            return True
        self._prefs_dialog = nfoview.PreferencesDialog(self)
        self._prefs_dialog.connect("close-request", on_close_request)
        self._prefs_dialog.show()

    def _on_wrap_lines_activate(self, action, *args):
        action.set_state(not action.get_state())
        if action.get_state():
            return self.view.set_wrap_mode(Gtk.WrapMode.WORD)
        return self.view.set_wrap_mode(Gtk.WrapMode.NONE)

    def open_file(self, path):
        if path is None: return
        self.path = Path(path).resolve()
        self.set_title(self.path.name)
        text = self._read_file(self.path)
        self.view.set_text(text)
        self.view.grab_focus()
        self._update_actions_enabled()

    def _read_file(self, path):
        encoding = nfoview.util.detect_encoding(path)
        lines = Path(path).read_text(encoding).splitlines()
        lines = [x.rstrip() for x in lines]
        while lines and not lines[-1]: lines.pop()
        # Handle erroneous (?) UTF-16 encoded files that use
        # NULL-character filled linebreaks '\x00\r\x00\n', which
        # readlines interprets as two separate linebreaks.
        if not any(lines[i] for i in range(1, len(lines), 2)):
            lines = [lines[i] for i in range(0, len(lines), 2)]
        return "\n".join(lines)

    def resize_to_text(self):
        # If the width of text exceeds 'text_view_max_chars',
        # switch to line wrapping and use 80 characters width.
        # Limit height to 'text_view_max_lines'. Finally limit
        # width and height to 80% of the screen.
        text = self.view.get_text()
        text = text or "\n".join(["x" * 80] * 40)
        size = list(nfoview.util.get_text_view_size(text))
        max_size = nfoview.util.get_max_text_view_size()
        if size[0] > max_size[0]:
            self.activate_action("wrap-lines", None)
            text = "\n".join(map(
                lambda x: textwrap.fill(x, 80),
                text.split("\n")))
            size = list(nfoview.util.get_text_view_size(text))
        size[0] = min(size[0], max_size[0])
        size[1] = min(size[1], max_size[1])
        # Assume 12 pixels total for text view margins,
        # 24 pixels for scrollbars, 48 pixels for header bar.
        size[0] = max(400, size[0] + 12 + 24)
        size[1] = max(248, size[1] + 12 + 24 + 48)
        screen_size = nfoview.util.get_screen_size()
        size[0] = min(size[0], int(0.8 * screen_size[0]))
        size[1] = min(size[1], int(0.8 * screen_size[1]))
        self.set_default_size(*size)

    def _update_actions_enabled(self):
        for name in self.list_actions():
            action = self.lookup_action(name)
            action.update_enabled(self)
